# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import copy
import shlex
import unittest

import pytest

from duplicity import (
    cli_main,
    gpg,
)
from duplicity.cli_data import *
from duplicity.cli_util import *
from testing import _runtest_dir
from testing.unit import UnitTestCase


@unittest.skipIf(os.environ.get("USER", "") == "buildd", "Skip test on Launchpad")
class CommandlineTest(UnitTestCase):
    """
    Test parse_commandline_options
    """

    good_args = {
        "count": "5",
        "remove_time": "100",
        "source_path": "foo/bar",
        "source_url": "file://duptest",
        "target_dir": "foo/bar",
        "target_url": "file://duptest",
    }

    def setUp(self):
        super().setUp()
        log.setup()
        config.gpg_profile = gpg.GPGProfile()
        os.makedirs("foo/bar", exist_ok=True)
        os.makedirs("inc", exist_ok=True)
        os.makedirs("full", exist_ok=True)

    def tearDown(self):
        log.shutdown()
        super().tearDown()

    def run_all_commands_with_errors(self, new_args, err_msg):
        """
        Test all commands with the supplied argument list.
        Only test command if new_args contains needed arg.
        """
        test_args = copy.copy(self.good_args)
        test_args.update(new_args)
        for var in DuplicityCommands.__dict__.keys():
            if var.startswith("__"):
                continue
            cmd = var2cmd(var)
            runtest = False
            args = DuplicityCommands.__dict__[var]
            cline = [cmd]
            for arg in args:
                cline.append(test_args[arg])
                if arg in new_args:
                    runtest = True
            if runtest:
                with self.assertRaisesRegex(cli_main.CommandLineError, err_msg) as cm:
                    cli_main.process_command_line(cline)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_full_command(self):
        """
        test backup, restore, verify with explicit commands
        """
        for cmd in ["cleanup"] + cli_main.CommandAliases.cleanup:
            cli_main.process_command_line(f"{cmd} file://duptest".split())
            self.assertEqual(config.action, "cleanup")
            self.assertEqual(config.target_url, "file://duptest")

        for cmd in ["collection-status"] + cli_main.CommandAliases.collection_status:
            cli_main.process_command_line(f"{cmd} file://duptest".split())
            self.assertEqual(config.action, "collection-status")
            self.assertEqual(config.target_url, "file://duptest")

        for cmd in ["full"] + cli_main.CommandAliases.full:
            cli_main.process_command_line(f"{cmd} foo/bar file://duptest".split())
            self.assertEqual(config.action, "full")
            self.assertTrue(config.source_path.endswith("foo/bar"))
            self.assertEqual(config.target_url, "file://duptest")

        for cmd in ["incremental"] + cli_main.CommandAliases.incremental:
            cli_main.process_command_line(f"{cmd} foo/bar file://duptest".split())
            self.assertEqual(config.action, "inc")
            self.assertTrue(config.source_path.endswith("foo/bar"))
            self.assertEqual(config.target_url, "file://duptest")

        for cmd in ["list-current-files"] + cli_main.CommandAliases.list_current_files:
            cli_main.process_command_line(f"{cmd} file://duptest".split())
            self.assertEqual(config.action, "list-current-files")
            self.assertEqual(config.target_url, "file://duptest")

        for cmd in ["remove-all-but-n-full"] + cli_main.CommandAliases.remove_all_but_n_full:
            cli_main.process_command_line(f"{cmd} 5 file://duptest".split())
            self.assertEqual(config.action, "remove-all-but-n-full")
            self.assertEqual(config.target_url, "file://duptest")

        for cmd in ["remove-all-inc-of-but-n-full"] + cli_main.CommandAliases.remove_all_inc_of_but_n_full:
            cli_main.process_command_line(f"{cmd} 5 file://duptest".split())
            self.assertEqual(config.action, "remove-all-inc-of-but-n-full")
            self.assertEqual(config.target_url, "file://duptest")

        for cmd in ["remove-older-than"] + cli_main.CommandAliases.remove_older_than:
            cli_main.process_command_line(f"{cmd} 100 file://duptest".split())
            self.assertEqual(config.action, "remove-older-than")
            self.assertEqual(config.target_url, "file://duptest")

        for cmd in ["restore"] + cli_main.CommandAliases.restore:
            cli_main.process_command_line(f"{cmd} file://duptest foo/bar".split())
            self.assertEqual(config.action, "restore")
            self.assertTrue(config.source_path.endswith("foo/bar"))
            self.assertEqual(config.target_url, "file://duptest")

        for cmd in ["verify"] + cli_main.CommandAliases.verify:
            cli_main.process_command_line(f"{cmd} file://duptest foo/bar".split())
            self.assertEqual(config.action, "verify")
            self.assertTrue(config.source_path.endswith("foo/bar"))
            self.assertEqual(config.target_url, "file://duptest")

    @pytest.mark.usefixtures("redirect_stdin")
    def test_full_command_errors_reversed_args(self):
        """
        test backup, restore, verify with explicit commands - reversed arg
        """
        new_args = {
            "source_path": "file://duptest",
            "source_url": "foo/bar",
            "target_dir": "file://duptest",
            "target_url": "foo/bar",
        }
        err_msg = "should be url|should be pathname"
        self.run_all_commands_with_errors(new_args, err_msg)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_full_command_errors_bad_url(self):
        """
        test backup, restore, verify with explicit commands - bad url
        """
        new_args = {
            "source_url": "file:/duptest",
            "target_url": "file:/duptest",
        }
        err_msg = "should be url"
        self.run_all_commands_with_errors(new_args, err_msg)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_full_command_errors_bad_integer(self):
        """
        test backup, restore, verify with explicit commands - bad integer
        """
        new_args = {
            "count": "foo",
        }
        err_msg = "not an int"
        self.run_all_commands_with_errors(new_args, err_msg)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_full_command_errors_bad_time_string(self):
        """
        test backup, restore, verify with explicit commands - bad time string
        """
        new_args = {
            "remove_time": "foo",
        }
        err_msg = "Bad time string"
        self.run_all_commands_with_errors(new_args, err_msg)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_option_aliases(self):
        """
        test short option aliases
        """
        cline = "ib foo/bar file:///target_url -v 9".split()
        cli_main.process_command_line(cline)
        self.assertEqual(log.getverbosity(), log.DEBUG)

        cline = "rb file:///source_url foo/bar -t 10000".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.restore_time, 10000)

        cline = "rb file:///source_url foo/bar --time 10000".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.restore_time, 10000)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_encryption_options(self):
        """
        test encrypt/sign key handling
        """
        start = "ib foo/bar file:///target_url "
        keys = (
            "DEADDEAD",
            "DEADDEADDEADDEAD",
            "DEADDEADDEADDEADDEADDEADDEADDEADDEADDEAD",
        )

        for key in keys:
            cline = f"{start} --encrypt-key={key}".split()
            cli_main.process_command_line(cline)
            self.assertEqual(config.gpg_profile.recipients, [key])

            cline = f"{start} --encrypt-sign-key={key}".split()
            cli_main.process_command_line(cline)
            self.assertEqual(config.gpg_profile.recipients, [key])
            self.assertEqual(config.gpg_profile.sign_key, key)

            cline = f"{start} --hidden-encrypt-key={key}".split()
            cli_main.process_command_line(cline)
            self.assertEqual(config.gpg_profile.hidden_recipients, [key])

            cline = f"{start} --sign-key={key}".split()
            cli_main.process_command_line(cline)
            self.assertEqual(config.gpg_profile.sign_key, key)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_bad_encryption_options(self):
        """
        test short option aliases
        """
        start = "inc foo/bar file:///target_url "
        keys = (
            "DEADFOO",
            "DEADDEADDEADFOO",
            "DEADDEADDEADDEADDEADDEADDEADDEADDEADFOO",
        )

        for key in keys:
            with self.assertRaises(CommandLineError) as cm:
                cline = f"{start} --encrypt-key={key}".split()
                cli_main.process_command_line(cline)

            with self.assertRaises(CommandLineError) as cm:
                cline = f"{start} --hidden-encrypt-key={key}".split()
                cli_main.process_command_line(cline)

            with self.assertRaises(CommandLineError) as cm:
                cline = f"{start} --sign-key={key}".split()
                cli_main.process_command_line(cline)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_implied_commands(self):
        """
        test implied commands
        """
        cline = "foo/bar file:///target_url".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "inc")
        self.assertEqual(config.source_path, "foo/bar")
        self.assertEqual(config.target_url, "file:///target_url")

        cline = "file:///source_url foo/bar".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "restore")
        self.assertEqual(config.source_url, "file:///source_url")
        self.assertEqual(config.target_dir, "foo/bar")

        cline = "-v9 foo/bar file:///target_url".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "inc")
        self.assertEqual(config.source_path, "foo/bar")
        self.assertEqual(config.target_url, "file:///target_url")

        cline = "-v9 file:///source_url foo/bar".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "restore")
        self.assertEqual(config.source_url, "file:///source_url")
        self.assertEqual(config.target_dir, "foo/bar")

        cline = "foo/bar -v9 file:///target_url".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "inc")
        self.assertEqual(config.source_path, "foo/bar")
        self.assertEqual(config.target_url, "file:///target_url")

        cline = "file:///source_url -v9 foo/bar".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "restore")
        self.assertEqual(config.source_url, "file:///source_url")
        self.assertEqual(config.target_dir, "foo/bar")

        cline = "--verbosity n foo/bar file:///target_url".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "inc")
        self.assertEqual(config.source_path, "foo/bar")
        self.assertEqual(config.target_url, "file:///target_url")

        cline = "--verbosity n file:///source_url foo/bar".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "restore")
        self.assertEqual(config.source_url, "file:///source_url")
        self.assertEqual(config.target_dir, "foo/bar")

        cline = "foo/bar --verbosity n file:///target_url".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "inc")
        self.assertEqual(config.source_path, "foo/bar")
        self.assertEqual(config.target_url, "file:///target_url")

        cline = "file:///source_url --verbosity n foo/bar".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "restore")
        self.assertEqual(config.source_url, "file:///source_url")
        self.assertEqual(config.target_dir, "foo/bar")

        # this incremental misses the path argument
        with self.assertRaises(CommandLineError) as cm:
            cline = "inc file:///target_url".split()
            cli_main.process_command_line(cline)

        # this full backup lacks the path argument
        with self.assertRaises(CommandLineError) as cm:
            cline = "full file:///target_url".split()
            cli_main.process_command_line(cline)

        # implied inc works if '/' supplied
        cline = "inc/ file:///target_url".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "inc")
        self.assertEqual(config.source_path, "inc/")
        self.assertEqual(config.target_url, "file:///target_url")

    @pytest.mark.usefixtures("redirect_stdin")
    def test_miscellaneous(self):
        """
        test miscellaneous parameters
        """
        start = "ib foo/bar file:///target_url"

        # check defaults, might add more asserts here
        cline = start.split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.print_statistics, True)

        cline = f"{start} --no-print-statistics".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.print_statistics, False)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_integer_args(self):
        """
        test implied commands
        """
        cline = "foo/bar file:///target_url --copy-blocksize=1024 --volsize=1024".split()
        cli_main.process_command_line(cline)
        self.assertEqual(config.copy_blocksize, 1024 * 1024)
        self.assertEqual(config.volsize, 1024 * 1024 * 1024)

        with self.assertRaises(CommandLineError) as cm:
            cline = "foo/bar file:///target_url --copy-blocksize=foo --volsize=1024".split()
            cli_main.process_command_line(cline)

        with self.assertRaises(CommandLineError) as cm:
            cline = "foo/bar file:///target_url --copy-blocksize=1024 --volsize=foo".split()
            cli_main.process_command_line(cline)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_bad_command(self):
        """
        test bad commands
        """
        with self.assertRaises(CommandLineError) as cm:
            cline = "fbx foo/bar file:///target_url".split()
            cli_main.process_command_line(cline)

        with self.assertRaises(CommandLineError) as cm:
            cline = "rbx file:///target_url foo/bar".split()
            cli_main.process_command_line(cline)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_too_many_positionals(self):
        """
        test bad commands
        """
        with self.assertRaises(CommandLineError) as cm:
            cline = "fb foo/bar file:///target_url extra".split()
            cli_main.process_command_line(cline)

        with self.assertRaises(CommandLineError) as cm:
            cline = "rb file:///target_url foo/bar extra".split()
            cli_main.process_command_line(cline)

        with self.assertRaises(CommandLineError) as cm:
            cline = "foo/bar file:///target_url extra".split()
            cli_main.process_command_line(cline)

        with self.assertRaises(CommandLineError) as cm:
            cline = "file:///target_url foo/bar extra".split()
            cli_main.process_command_line(cline)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_list_commands(self):
        """
        test list commands like ssh_options, etc.
        """
        cline = shlex.split("inc foo/bar file:///target_url --ssh-options='--foo'")
        cli_main.process_command_line(cline)
        self.assertEqual(config.ssh_options, "--foo")

        cline = shlex.split("inc foo/bar file:///target_url --ssh-options='--foo' --ssh-options='--bar'")
        cli_main.process_command_line(cline)
        self.assertEqual(config.ssh_options, "--foo --bar")

        cline = shlex.split("inc foo/bar file:///target_url --ssh-options='--foo --bar'")
        cli_main.process_command_line(cline)
        self.assertEqual(config.ssh_options, "--foo --bar")

    @pytest.mark.usefixtures("redirect_stdin")
    def test_help_commands(self):
        """
        Test -h/--help
        """
        with self.assertRaises(SystemExit) as cm:
            cli_main.process_command_line(shlex.split("-h"))
            self.assertTrue(check_main_help(cm.content))
        with self.assertRaises(SystemExit) as cm:
            cli_main.process_command_line(shlex.split(f"--help"))
            self.assertTrue(check_main_help(cm.content))

        for cmd in [var2cmd(v) for v in DuplicityCommands.__dict__.keys() if not v.startswith("__")]:
            with self.assertRaises(SystemExit) as cm:
                cli_main.process_command_line(shlex.split(f"{cmd} -h"))
            with self.assertRaises(SystemExit) as cm:
                cli_main.process_command_line(shlex.split(f"{cmd} --help"))

    @pytest.mark.usefixtures("redirect_stdin")
    def test_log_options(self):
        """
        test log options.
        """
        log.setup()
        # TODO: this fails although running duplicity, cli_main return the correct default loglevel
        # default level is notice
        # self.assertEqual(log.getverbosity(), log.NOTICE)

        # setting custom level
        cline = shlex.split("foo/bar file:///target_url --verbosity Debug")
        cli_main.process_command_line(cline)
        self.assertEqual(log.getverbosity(), log.DEBUG)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_changed_removed(self):
        """
        test changed/removed in odd places.
        """
        # removed option with command
        with self.assertRaises(cli_main.CommandLineError) as cm:
            cline = shlex.split("--gio backup foo/bar file:///target_url")
            cli_main.process_command_line(cline)

        # removed option without command
        with self.assertRaises(CommandLineError) as cm:
            cline = shlex.split("--gio foo/bar file:///target_url")
            cli_main.process_command_line(cline)

        # changed option with command
        with self.assertRaises(CommandLineError) as cm:
            cline = shlex.split("restore --file-to-restore foo/bar file://source_url path")
            cli_main.process_command_line(cline)

        # changed option without command
        with self.assertRaises(CommandLineError) as cm:
            cline = shlex.split("--file-to-restore foo/bar file://source_url path")
            cli_main.process_command_line(cline)

        # removed backup option with command
        with self.assertRaises(CommandLineError) as cm:
            cline = shlex.split("--time-separator _ backup source_dir file://target_url")
            cli_main.process_command_line(cline)

        # removed backup option without command
        with self.assertRaises(CommandLineError) as cm:
            cline = shlex.split("--time-separator _ source_dir file://target_url")
            cli_main.process_command_line(cline)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_intermixed_args(self):
        """
        test intermixed args.
        """
        # Issue 766 -- intermixed -- explicit
        cline = shlex.split(
            f"--archive-dir {_runtest_dir}/backup-metadata/archive/ --tempdir {_runtest_dir}/backup-metadata/temp/ "
            f"--allow-source-mismatch --encrypt-sign-key DEADDEAD --volsize 4096 --progress -v 4 "
            f"incr --full-if-older-than 30D foo/bar --log-file {_runtest_dir}/log.txt boto3+s3://foo"
        )
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "inc")
        self.assertEqual(config.allow_source_mismatch, True)
        self.assertEqual(config.archive_dir, f"{_runtest_dir}/backup-metadata/archive/".encode())
        self.assertEqual(config.full_if_older_than, 2592000)
        self.assertEqual(config.progress, True)
        self.assertEqual(config.temproot, f"{_runtest_dir}/backup-metadata/temp/".encode())
        self.assertEqual(config.volsize, 4294967296)

        # Issue 766 -- intermixed -- implicit
        cline = shlex.split(
            f"--archive-dir {_runtest_dir}/backup-metadata/archive/ --tempdir {_runtest_dir}/backup-metadata/temp/ "
            f"--allow-source-mismatch --encrypt-sign-key DEADDEAD --volsize 4096 --progress -v 4 "
            f"--full-if-older-than 30D foo/bar --log-file {_runtest_dir}/log.txt boto3+s3://foo"
        )
        cli_main.process_command_line(cline)
        self.assertEqual(config.action, "inc")
        self.assertEqual(config.allow_source_mismatch, True)
        self.assertEqual(config.archive_dir, f"{_runtest_dir}/backup-metadata/archive/".encode())
        self.assertEqual(config.full_if_older_than, 2592000)
        self.assertEqual(config.progress, True)
        self.assertEqual(config.temproot, f"{_runtest_dir}/backup-metadata/temp/".encode())
        self.assertEqual(config.volsize, 4294967296)

    @pytest.mark.usefixtures("redirect_stdin")
    def test_regression_issues(self):
        """
        test regression issues.
        """
        # Issue 759 - --asynchronous-upload
        cline = shlex.split("backup --asynchronous-upload / file://target_url")
        cli_main.process_command_line(cline)
        self.assertEqual(config.async_concurrency, 1)

        # Issue 764 - --full-if-older-than -- explicit
        cline = shlex.split("backup --full-if-older-than 30D foo/bar file://target_url")
        cli_main.process_command_line(cline)
        self.assertEqual(config.full_if_older_than, 2592000)

        # Issue 764 - --full-if-older-than -- implied
        cline = shlex.split("--full-if-older-than 30D foo/bar file://target_url")
        cli_main.process_command_line(cline)
        self.assertEqual(config.full_if_older_than, 2592000)

        # Issue 773 - --exclude-device-files
        config.select_opts = []
        cline = shlex.split("backup --exclude-device-files / file://target_url")
        cli_main.process_command_line(cline)
        self.assertListEqual(config.select_opts, [("--exclude-device-files", [])])

        # Issue 773 - --exclude-other-fileystems
        config.select_opts = []
        cline = shlex.split("backup --exclude-other-filesystems / file://target_url")
        cli_main.process_command_line(cline)
        self.assertListEqual(config.select_opts, [("--exclude-other-filesystems", [])])

        # Issue 795/816 - invalid option error using --gpg-options - unbound - argparse bug
        with self.assertRaises(CommandLineError) as cm:
            cline = shlex.split("backup --gpg-options '--homedir=/home/user' foo/bar file://target_url")
            cli_main.process_command_line(cline)
            self.assertIn("design error in argparse", cm.exception)

        # Issue 795/816 - invalid option error using --gpg-options - bound
        cline = shlex.split("backup --gpg-options='--homedir=/home/user' foo/bar file://target_url")
        cli_main.process_command_line(cline)
        self.assertEqual(config.gpg_options, "--homedir=/home/user")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import unittest

from duplicity.path import *  # pylint: disable=unused-wildcard-import,redefined-builtin
from testing import _runtest_dir
from . import UnitTestCase


class PathTest(UnitTestCase):
    """Test basic path functions"""

    def setUp(self):
        super().setUp()
        self.unpack_testfiles()

    def test_deltree(self):
        """Test deleting a tree"""
        assert not os.system(f"cp -pR {_runtest_dir}/testfiles/deltree {_runtest_dir}/testfiles/output")
        p = Path(f"{_runtest_dir}/testfiles/output")
        assert p.isdir()
        p.deltree()
        assert not p.type, p.type

    def test_quote(self):
        """Test path quoting"""
        p = Path("hello")
        assert p.quote() == '"hello"'
        assert p.quote("\\") == '"\\\\"', p.quote("\\")
        assert p.quote("$HELLO") == '"\\$HELLO"'

    def test_unquote(self):
        """Test path unquoting"""
        p = Path("foo")  # just to provide unquote function

        def t(s):
            """Run test on string s"""
            quoted_version = p.quote(s)
            unquoted = p.unquote(quoted_version)
            assert unquoted == s, (unquoted, s)

        t("\\")
        t("$HELLO")
        t(" aoe aoe \\ \n`")

    def test_canonical(self):
        """Test getting canonical version of path"""
        c = Path(".").get_canonical()
        assert c == b".", c

        c = Path("//foo/bar/./").get_canonical()
        assert c == b"/foo/bar", c

    def test_compare_verbose(self):
        """Run compare_verbose on a few files"""
        vft = Path(f"{_runtest_dir}/testfiles/various_file_types")
        assert vft.compare_verbose(vft)
        reg_file = vft.append("regular_file")
        assert not vft.compare_verbose(reg_file)
        assert reg_file.compare_verbose(reg_file)
        file2 = vft.append("executable")
        assert not file2.compare_verbose(reg_file)
        assert file2.compare_verbose(file2)


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2011 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import unittest
from duplicity import log

from duplicity.backends._testbackend import BackendErrors as BE

import pytest

from testing import _runtest_dir
from testing.functional import CmdError
from testing.functional import FunctionalTestCase


class BadUploadTestBackend(FunctionalTestCase):
    """
    Test missing volume upload using duplicity binary
    """

    def _put_fail_volume(self, failure, condition, cmderror):
        """
        _testbackend throw exception on put of certain volume
        """
        self.make_largefiles()
        self.backup_with_failure("full", f"{_runtest_dir}/testfiles/largefiles", failure, condition, cmderror)

    @pytest.mark.slow
    def test_skip_volume_silent(self):
        """
        _testbackend won't put a certain volume
        """
        self.backup_with_failure(
            "full",
            f"{_runtest_dir}/testfiles/dir1",
            BE.SKIP_PUT_SILENT,
            "vol1.difftar",
            log.ErrorCode.backend_validation_failed,
            # PYDEVD="vscode"
        )

    @pytest.mark.slow
    def test_put_fail_volume1(self):
        """
        _testbackend throw exception on put of singel volume
        """
        self.backup_with_failure(
            "full",
            f"{_runtest_dir}/testfiles/dir1",
            BE.FAIL_WITH_EXCEPTION,
            "vol1.difftar",
            log.ErrorCode.backend_error,
        )

    @pytest.mark.slow
    def test_put_fail_volume2(self):
        """
        _testbackend throw exception on put of volume2 of serverals
        """
        self.make_largefiles()
        self.backup_with_failure(
            "full",
            f"{_runtest_dir}/testfiles/largefiles",
            BE.FAIL_WITH_EXCEPTION,
            "vol2.difftar",
            log.ErrorCode.backend_error,
        )


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2012 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import gettext
import os
import platform
import subprocess
import sys
import time
import unittest
from importlib import reload

from duplicity import (
    backend,
    config,
    log,
    util,
)

gettext.install("duplicity", names=["ngettext"])

log.setup()
util.start_debugger()

_testing_dir = os.path.dirname(os.path.abspath(__file__))
_top_dir = os.path.dirname(_testing_dir)
_overrides_dir = os.path.join(_testing_dir, "overrides")
_bin_dir = os.path.join(_testing_dir, "overrides", "bin")

if platform.system().startswith("Darwin"):
    # Use temp space TMPDIR or from getconf, never /tmp
    _runtest_dir = os.environ.get("TMPDIR", None) or subprocess.check_output(["getconf", "DARWIN_USER_TEMP_DIR"])
    _runtest_dir = os.fsdecode(_runtest_dir).rstrip().rstrip("/")
    if not os.path.exists(_runtest_dir):
        os.makedirs(_runtest_dir)
else:
    # be a little more flexible
    _runtest_dir = os.getenv("TMPDIR", False) or os.getenv("TEMP", False) or "/tmp"

if not os.path.exists(_runtest_dir):
    os.makedirs(_runtest_dir)

# Adjust python path for duplicity and override modules
sys.path = [_overrides_dir, _top_dir, _bin_dir] + sys.path

# Also set PYTHONPATH for any subprocesses
os.environ["PYTHONPATH"] = f"{_overrides_dir}:{_top_dir}:{os.environ.get('PYTHONPATH', '')}"

# And PATH for any subprocesses
os.environ["PATH"] = f"{_bin_dir}:{os.environ.get('PATH', '')}"

# Now set some variables that help standardize test behavior
os.environ["LANG"] = ""

# Set up GNUPGHOME for testing on system or docker
if os.environ.get("DOCKER_GNUPGHOME", ""):
    os.environ["GNUPGHOME"] = os.environ["DOCKER_GNUPGHOME"]
else:
    os.environ["GNUPGHOME"] = os.path.join(_testing_dir, "gnupg")

# bzr does not honor perms so fix the perms and avoid annoying error
os.system(f"chmod 700 {os.path.join(_testing_dir, 'gnupg')}")

# Standardize time
os.environ["TZ"] = "US/Central"
time.tzset()


class DuplicityTestCase(unittest.TestCase):
    sign_key = "839E6A2856538CCF"
    sign_passphrase = "test"
    encrypt_key1 = "839E6A2856538CCF"
    encrypt_key2 = "453005CE9B736B2A"

    def setUp(self):
        super().setUp()
        self.savedEnviron = {}
        self.savedConfig = {}

        log.setup()
        log.setverbosity(log.WARNING)
        self.set_config("print_statistics", 0)
        backend.import_backends()

        self.remove_testfiles()
        self.unpack_testfiles()

        self.set_environ("TZ", "UTC")
        time.tzset()
        assert time.tzname[0] == "UTC", f"{time.tzname[0]} should be 'UTC'"

        # Have all file references in tests relative to our runtest dir
        os.chdir(_runtest_dir)

        # reimport duplicity.config in case it changed
        reload(config)

    def tearDown(self):
        for key in self.savedEnviron:
            self._update_env(key, self.savedEnviron[key])

        for key in self.savedConfig:
            setattr(config, key, self.savedConfig[key])

        time.tzset()

        self.remove_testfiles()

        os.chdir(_testing_dir)
        super().tearDown()

    def unpack_testfiles(self):
        assert not os.system(f"rm -rf {_runtest_dir}/testfiles")
        assert not os.system(f"tar xzf {_testing_dir}/testfiles.tar.gz -C {_runtest_dir} > /dev/null 2>&1")
        assert not os.system(f"mkdir {_runtest_dir}/testfiles/output {_runtest_dir}/testfiles/cache")

    def remove_testfiles(self):
        assert not os.system(f"rm -rf {_runtest_dir}/testfiles")

    def _update_env(self, key, value):
        if value is not None:
            os.environ[key] = value
        elif key in os.environ:
            del os.environ[key]

    def set_environ(self, key, value):
        if key not in self.savedEnviron:
            self.savedEnviron[key] = os.environ.get(key)
        self._update_env(key, value)

    def set_config(self, key, value):
        assert hasattr(config, key)
        if key not in self.savedConfig:
            self.savedConfig[key] = getattr(config, key)
        setattr(config, key, value)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#

import io
import os
import sys

import pytest

from testing import _runtest_dir


@pytest.fixture(scope="function")
def redirect_stdin():
    """
    GPG requires stdin to be open and have real file descriptor, which interferes with pytest's capture facility.
    Work around this by redirecting /dev/null to stdin temporarily.

    Activate this fixture on unittest test methods and classes by means of:
    @pytest.mark.usefixtures("redirect_stdin").
    """
    try:
        targetfd_save = os.dup(0)
        stdin_save = sys.stdin

        nullfile = open(os.devnull, "r")
        sys.stdin = nullfile
        os.dup2(nullfile.fileno(), 0)
        yield
    finally:
        os.dup2(targetfd_save, 0)  # pylint: disable=used-before-assignment
        sys.stdin = stdin_save  # pylint: disable=used-before-assignment
        os.close(targetfd_save)
        nullfile.close()  # pylint: disable=used-before-assignment


@pytest.fixture(scope="function")
def fake_input():
    orig_stdin = sys.stdin
    sys.stdin = io.StringIO("yes\n")
    yield
    sys.stdin = orig_stdin


@pytest.hookimpl()
def pytest_sessionfinish(session, exitstatus):
    """
    Called after whole test run finished, right before
    returning the exit status to the system.  Since tests
    kill duplicity and cause errors intentionally this is
    necessary to keep a clean test system.
    """
    cleanup = [
        "backup-metadata",
        "duplicity-*-tempdir",
        "duptest",
        "foo",
        "full",
        "inc",
        "log.txt",
        "target_url",
        "testbackend.log",
    ]
    for fn in cleanup:
        os.system(f"rm -rf {_runtest_dir}/{fn}")



#!/usr/bin/env python3

import glob
import json
import os
import re
from pprint import pprint as print
import matplotlib.pyplot as plt


DEST_DIR = "/tmp/back-cc-test"

stats = {}
rex = re.compile(r".*tt([0-9]*)-cc([0-9]*).json")

for filename in glob.iglob("/tmp/results-*/jsonstat-tt*.json"):
    print(filename)
    m = rex.match(filename)
    tt, cc = m.groups() if m is not None else ("no", "no")
    tt, cc = int(tt), int(cc)
    with open(filename) as file:
        stats.setdefault(tt, {})[cc] = json.load(file)
        pass
# print(stats)

for tt, cc_dict in sorted(stats.items()):
    time_per_volumes = []
    for cc, values in cc_dict.items():
        time_total = values["ElapsedTime"]
        time_per_volume = time_total / values["ConcurrentTransferStats"]["count"]
        time_per_volumes.append(time_per_volume)
        print(
            f"Backend delay: {tt:5}, Concurrency: {cc} Time: {values['ElapsedTime']:6.2f}, Time per vol: {time_per_volume:5.2f}, avg pur transfer time {values['ConcurrentTransferStats']['time']['avg']:5.2f}, avg throttle {values['ConcurrentTransferStats']['throttle']['avg']:5.2f}, Peak in queue: {values['ConcurrentTransferStats']['peak_in_queue']}",
            width=200,
        )
    x1 = list(sorted(cc_dict.keys()))
    y1 = [y["ElapsedTime"] for y in cc_dict.values()]
    y1 = time_per_volumes
    plt.plot(x1, y1, label=f"TT: {tt}")

# naming the x axis
plt.xlabel("concurrency")
# naming the y axis
plt.ylabel("avg duration per volume")
# giving a title to my graph
plt.title("Backup time by concurrency and transfer speed")

# show a legend on the plot
plt.legend()

# function to show the plot
plt.savefig("/tmp/concurrency-stat.png")
plt.show()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2014 Michael Terry <michael.terry@canonical.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import glob
import os
import pycodestyle
import sys
from subprocess import (
    Popen,
    PIPE,
    STDOUT,
)
import unittest

import pytest

from . import _top_dir, DuplicityTestCase

files_to_test = []
files_to_test.extend(glob.glob(os.path.join(_top_dir, "duplicity/**/*.py"), recursive=True))
files_to_test.extend(glob.glob(os.path.join(_top_dir, "testing/functional/*.py")))
files_to_test.extend(glob.glob(os.path.join(_top_dir, "testing/unit/*.py")))
files_to_test.extend(glob.glob(os.path.join(_top_dir, "testing/*.py")))

# don't test argparse311.py.  not really ours.
files_to_test.remove(os.path.join(_top_dir, "duplicity/argparse311.py"))

# TODO: remove when pylint AST builder is fixed.
files_to_test.remove(os.path.join(_top_dir, "duplicity/backends/giobackend.py"))


@unittest.skipIf(os.environ.get("USER", "") == "buildd", "Skip test on Launchpad")
class CodeTest(DuplicityTestCase):
    def run_checker(self, cmd, returncodes=None):
        if returncodes is None:
            returncodes = [0]
        process = Popen(cmd, stdout=PIPE, stderr=STDOUT, universal_newlines=True)
        output = process.communicate()[0]
        if len(output):
            for line in output.split("\n"):
                print(line, file=sys.stderr)
            output = ""
        self.assertTrue(
            process.returncode in returncodes,
            f"Test failed: returncode = {process.returncode}",
        )

    def test_black(self):
        """Black check for out of format files"""
        print()
        self.run_checker(
            [
                "black",
                "--check",
            ]
            + files_to_test,
        )

    def test_pep8(self):
        """Test that we conform to PEP-8 using pycodestyle."""
        # Note that the settings, ignores etc for pycodestyle are set in pyproject.toml, not here
        print()
        style = pycodestyle.StyleGuide(config_file=os.path.join(_top_dir, "setup.cfg"))
        result = style.check_files(files_to_test)
        self.assertEqual(
            result.total_errors,
            0,
            f"Found {result.total_errors} code style errors (and warnings).",
        )

    def test_pylint(self):
        """Pylint test (requires pylint to be installed to pass)"""
        print()
        self.run_checker(
            [
                "pylint",
                f"--rcfile={os.path.join(_top_dir, 'pyproject.toml')}",
            ]
            + files_to_test
        )


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import os
import unittest

import pytest

from duplicity import path
from testing.functional import (
    _runtest_dir,
    CmdError,
    FunctionalTestCase,
)


class FinalTest(FunctionalTestCase):
    """
    Test backup/restore using duplicity binary
    """

    def runtest(self, dirlist, backup_options=None, restore_options=None):
        """Run backup/restore test on directories in dirlist"""
        if backup_options is None:
            backup_options = []
        if restore_options is None:
            restore_options = []

        assert len(dirlist) >= 1

        backup_options += ["--allow-source-mismatch"]

        # Back up directories to local backend
        current_time = 100000
        self.backup("full", dirlist[0], current_time=current_time, options=backup_options)
        for new_dir in dirlist[1:]:
            current_time += 100000
            self.backup("inc", new_dir, current_time=current_time, options=backup_options)

        # Restore each and compare them
        for i in range(len(dirlist)):
            dirname = dirlist[i]
            current_time = 100000 * (i + 1)
            self.restore(time=current_time, options=restore_options)
            self.check_same(dirname, f"{_runtest_dir}/testfiles/restore_out")
            self.verify(dirname, time=current_time, options=restore_options)

    def check_same(self, filename1, filename2):
        """Verify two filenames are the same"""
        path1, path2 = path.Path(filename1), path.Path(filename2)
        assert path1.compare_recursive(path2, verbose=1)

    @pytest.mark.slow
    def test_basic_cycle(self, backup_options=None, restore_options=None, dirlist=None, testfiles=None):
        """Run backup/restore test on basic directories"""
        if backup_options is None:
            backup_options = ["--no-encrypt", "--no-compress"]
        if restore_options is None:
            restore_options = ["--no-encrypt", "--no-compress"]
        if dirlist is None:
            dirlist = [
                f"{_runtest_dir}/testfiles/dir1",
                f"{_runtest_dir}/testfiles/dir2",
                f"{_runtest_dir}/testfiles/dir3",
            ]
        self.runtest(dirlist, backup_options=backup_options, restore_options=restore_options)

        if testfiles is None:
            testfiles = [
                ("symbolic_link", 99999, "dir1"),
                ("directory_to_file", 100100, "dir1"),
                ("directory_to_file", 200100, "dir2"),
                ("largefile", 300000, "dir3"),
            ]
        # Test restoring various sub files
        for filename, time, tfdir in testfiles:
            self.restore(filename, time, options=restore_options)
            self.check_same(f"{_runtest_dir}/testfiles/{tfdir}/{filename}", f"{_runtest_dir}/testfiles/restore_out")
            self.verify(
                f"{_runtest_dir}/testfiles/{tfdir}/{filename}",
                file_to_verify=filename,
                time=time,
                options=restore_options,
            )

    @pytest.mark.slow
    def test_asym_cycle(self):
        """Like test_basic_cycle but use asymmetric encryption and signing"""
        backup_options = ["--encrypt-key", self.encrypt_key1, "--sign-key", self.sign_key]
        restore_options = ["--encrypt-key", self.encrypt_key1, "--sign-key", self.sign_key]
        self.test_basic_cycle(backup_options=backup_options, restore_options=restore_options)

    @pytest.mark.slow
    def test_asym_with_hidden_recipient_cycle(self):
        """Like test_basic_cycle but use asymmetric encryption (hiding key id) and signing"""
        backup_options = ["--hidden-encrypt-key", self.encrypt_key1, "--sign-key", self.sign_key]
        restore_options = ["--hidden-encrypt-key", self.encrypt_key1, "--sign-key", self.sign_key]
        self.test_basic_cycle(backup_options=backup_options, restore_options=restore_options)

    def test_single_regfile(self):
        """Test backing and restoring up a single regular file"""
        self.runtest([f"{_runtest_dir}/testfiles/various_file_types/regular_file"])

    def test_empty_backup(self):
        """Make sure backup works when no files change"""
        self.backup("full", f"{_runtest_dir}/testfiles/empty_dir")
        self.backup("inc", f"{_runtest_dir}/testfiles/empty_dir")

    @pytest.mark.slow
    def test_long_filenames(self):
        """Test backing up a directory with long filenames in it"""
        # Note that some versions of ecryptfs (at least through Ubuntu 11.10)
        # have a bug where they treat the max path segment length as 143
        # instead of 255.  So make sure that these segments don't break that.
        lf_dir = path.Path(f"{_runtest_dir}/testfiles/long_filenames")
        if lf_dir.exists():
            lf_dir.deltree()
        lf_dir.mkdir()
        lf1 = lf_dir.append(
            "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"  # noqa
        )
        lf1.mkdir()
        lf2 = lf1.append(
            "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB"  # noqa
        )
        lf2.mkdir()
        lf3 = lf2.append(
            "CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC"  # noqa
        )
        lf3.mkdir()
        lf4 = lf3.append(
            "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD"  # noqa
        )
        lf4.touch()
        lf4_1 = lf3.append(
            "SYMLINK--------------------------------------------------------------------------------------------"  # noqa
        )
        os.symlink(
            "SYMLINK-DESTINATION-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------",  # noqa
            lf4_1.name,
        )
        lf4_1.setdata()
        assert lf4_1.issym()
        lf4_2 = lf3.append(
            "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD"  # noqa
        )
        fp = lf4_2.open("wb")
        fp.write(b"hello" * 1000)
        assert not fp.close()

        self.runtest(
            [
                f"{_runtest_dir}/testfiles/empty_dir",
                lf_dir.uc_name,
                f"{_runtest_dir}/testfiles/empty_dir",
                lf_dir.uc_name,
            ]
        )

    def test_empty_restore(self):
        """Make sure error raised when restore doesn't match anything"""
        self.backup("full", f"{_runtest_dir}/testfiles/dir1", options=["--allow-source-mismatch"])
        self.assertRaises(CmdError, self.restore, "this_file_does_not_exist")
        self.backup("inc", f"{_runtest_dir}/testfiles/empty_dir", options=["--allow-source-mismatch"])
        self.assertRaises(CmdError, self.restore, "this_file_does_not_exist")

    @pytest.mark.slow
    def test_remove_older_than(self):
        """Test removing old backup chains"""
        first_chain = self.backup(
            "full", f"{_runtest_dir}/testfiles/dir1", current_time=10000, options=["--allow-source-mismatch"]
        )
        first_chain |= self.backup(
            "inc", f"{_runtest_dir}/testfiles/dir2", current_time=20000, options=["--allow-source-mismatch"]
        )
        second_chain = self.backup(
            "full", f"{_runtest_dir}/testfiles/dir1", current_time=30000, options=["--allow-source-mismatch"]
        )
        second_chain |= self.backup(
            "inc", f"{_runtest_dir}/testfiles/dir3", current_time=40000, options=["--allow-source-mismatch"]
        )

        self.assertEqual(self.get_backend_files(), first_chain | second_chain)

        self.run_duplicity(options=["remove-older-than", "35000", "--force", self.backend_url])
        self.assertEqual(self.get_backend_files(), second_chain)

        # Now check to make sure we can't delete only chain
        self.run_duplicity(options=["remove-older-than", "50000", "--force", self.backend_url])
        self.assertEqual(self.get_backend_files(), second_chain)

    def test_piped_password(self):
        """Make sure that prompting for a password works"""
        self.set_environ("PASSPHRASE", None)
        self.backup(
            "full", f"{_runtest_dir}/testfiles/empty_dir", passphrase_input=[self.sign_passphrase, self.sign_passphrase]
        )
        self.restore(passphrase_input=[self.sign_passphrase])

    @pytest.mark.slow
    def test_jsonstat(self):
        """Test cycle with json stats enabled"""
        backup_options = ["--jsonstat"]
        restore_options = ["--jsonstat"]
        self.test_basic_cycle(backup_options=backup_options, restore_options=restore_options)

    def test_jsonstat_missing(self):
        """Make sure collection_status works if one set misses jsonstat"""
        self.backup("full", f"{_runtest_dir}/testfiles/dir1", options=["--jsonstat", "--allow-source-mismatch"])
        self.backup("inc", f"{_runtest_dir}/testfiles/empty_dir", options=["--jsonstat", "--allow-source-mismatch"])
        self.backup("inc", f"{_runtest_dir}/testfiles/dir2", options=["--allow-source-mismatch"])
        self.collection_status(options=["--show-changes-in-set", "-1", "--jsonstat"])

    def run_with_no_change(self, backup_options=None, restore_options=None):
        """Like test_basic_cycle but runs a incremental backup with no changes after full"""
        testfiles = [
            ("symbolic_link", 99999, "dir1"),
            ("directory_to_file", 100100, "dir1"),
            ("directory_to_file", 300100, "dir2"),
            ("largefile", 400000, "dir3"),
        ]
        self.test_basic_cycle(
            backup_options=backup_options,
            restore_options=restore_options,
            dirlist=[
                f"{_runtest_dir}/testfiles/dir1",
                f"{_runtest_dir}/testfiles/dir1",
                f"{_runtest_dir}/testfiles/dir2",
                f"{_runtest_dir}/testfiles/dir3",
            ],
            testfiles=testfiles,
        )

    @pytest.mark.slow
    def test_skip_if_no_change(self):
        self.run_with_no_change(backup_options=["--skip-if-no-change"])

    @pytest.mark.slow
    def test_concurrency(self):
        backup_options = ["--concurrency=2"]
        self.test_basic_cycle(backup_options=backup_options)

    @pytest.mark.slow
    def test_concurrency_and_skip_if_no_change(self):
        backup_options = ["--concurrency=2", "--skip-if-no-change"]
        self.run_with_no_change(backup_options=backup_options)


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2014 Aaron Whitehouse <aaron@whitehouse.kiwi.nz>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import os
import unittest

from testing import _runtest_dir
from . import (
    CmdError,
    FunctionalTestCase,
)


class VerifyTest(FunctionalTestCase):
    """
    Test verify using duplicity binary
    """

    def test_verify(self):
        """Test that verify (without --compare-data) works in the basic case"""
        self.backup("full", f"{_runtest_dir}/testfiles/various_file_types", options=[])
        self.verify(
            f"{_runtest_dir}/testfiles/various_file_types/executable",
            file_to_verify="executable",
            options=[],
        )

    def test_verify_changed_source_file(self):
        """Test verify (without --compare-data) gives no error if a source file is changed"""
        # This test was made to pass in fixing Bug #1354880
        self.backup("full", f"{_runtest_dir}/testfiles/various_file_types", options=[])

        # Edit source file
        with open(f"{_runtest_dir}/testfiles/various_file_types/executable", "r+") as f:
            f.write("This changes a source file.")

        # Test verify for the file
        self.verify(
            f"{_runtest_dir}/testfiles/various_file_types/executable",
            file_to_verify="executable",
            options=[],
        )

    def test_verify_changed_source_file_adjust_mtime(self):
        """Test verify (without --compare-data) gives no error if a source file is changed and the mtime is changed
        (changing anything about the source files shouldn't matter)"""

        # Get the atime and mtime of the file
        file_info = os.stat(f"{_runtest_dir}/testfiles/various_file_types/executable")

        # Set the atime and mtime of the file to the time that we collected, as on some systems
        # the times from a stat call don't match what a utime will set.
        os.utime(
            f"{_runtest_dir}/testfiles/various_file_types/executable",
            (file_info.st_atime, file_info.st_mtime),
        )

        self.backup("full", f"{_runtest_dir}/testfiles/various_file_types", options=[])

        # Edit source file
        with open(f"{_runtest_dir}/testfiles/various_file_types/executable", "r+") as f:
            f.write("This changes a source file.")

        # Set the atime and mtime for the file back to what it was prior to the edit
        os.utime(
            f"{_runtest_dir}/testfiles/various_file_types/executable",
            (file_info.st_atime, file_info.st_mtime),
        )

        # Test verify for the file
        self.verify(
            f"{_runtest_dir}/testfiles/various_file_types/executable",
            file_to_verify="executable",
            options=[],
        )

    def test_verify_compare_data(self):
        """Test that verify works in the basic case when the --compare-data option is used"""
        self.backup("full", f"{_runtest_dir}/testfiles/various_file_types", options=[])

        # Test verify for the file with --compare-data
        self.verify(
            f"{_runtest_dir}/testfiles/various_file_types/executable",
            file_to_verify="executable",
            options=["--compare-data"],
        )

    def test_verify_compare_data_changed_source_file(self):
        """Test verify with --compare-data gives an error if a source file is changed"""
        self.backup("full", f"{_runtest_dir}/testfiles/various_file_types", options=[])

        # Edit source file
        with open(f"{_runtest_dir}/testfiles/various_file_types/executable", "r+") as f:
            f.write("This changes a source file.")

        # Test verify for edited file fails with --compare-data
        try:
            self.verify(
                f"{_runtest_dir}/testfiles/various_file_types/executable",
                file_to_verify="executable",
                options=["--compare-data"],
            )
        except CmdError as e:
            self.assertEqual(e.exit_status, 1, str(e))
        else:
            self.fail("Expected CmdError not thrown")

    def test_verify_compare_data_changed_source_file_adjust_mtime(self):
        """Test verify with --compare-data gives an error if a source file is changed, even if the mtime is changed"""

        # Get the atime and mtime of the file
        file_info = os.stat(f"{_runtest_dir}/testfiles/various_file_types/executable")

        # Set the atime and mtime of the file to the time that we collected, as on some systems
        # the times from a stat call don't match what a utime will set
        os.utime(
            f"{_runtest_dir}/testfiles/various_file_types/executable",
            (file_info.st_atime, file_info.st_mtime),
        )

        self.backup("full", f"{_runtest_dir}/testfiles/various_file_types", options=[])
        # Edit source file
        with open(f"{_runtest_dir}/testfiles/various_file_types/executable", "r+") as f:
            f.write("This changes a source file.")

        # Set the atime and mtime for the file back to what it was prior to the edit
        os.utime(
            f"{_runtest_dir}/testfiles/various_file_types/executable",
            (file_info.st_atime, file_info.st_mtime),
        )

        # Test verify for edited file fails with --compare-data
        try:
            self.verify(
                f"{_runtest_dir}/testfiles/various_file_types/executable",
                file_to_verify="executable",
                options=["--compare-data"],
            )
        except CmdError as e:
            self.assertEqual(e.exit_status, 1, str(e))
        else:
            self.fail("Expected CmdError not thrown")

    def test_verify_corrupt_archive(self):
        """Test verify (without --compare-data) gives an error if the archive is corrupted"""
        self.backup("full", f"{_runtest_dir}/testfiles/various_file_types", options=[])
        output_files = os.listdir(f"{_runtest_dir}/testfiles/output")
        archives = [elem for elem in output_files if "vol" in elem]
        for archive in archives:
            # Edit source file
            with open(f"{_runtest_dir}/testfiles/output/{archive}", "r+") as f:
                f.write("This writes text into each archive file to corrupt it.")
        # Test verify for the file
        try:
            self.verify(
                f"{_runtest_dir}/testfiles/various_file_types/executable",
                file_to_verify="executable",
                options=[],
            )
        except CmdError as e:
            # Should return a 21 error code for "hash mismatch"
            self.assertEqual(e.exit_status, 21, str(e))
        else:
            self.fail("Expected Hash Mismatch Error not thrown")

    def test_verify_corrupt_archive_compare_data(self):
        """Test verify with --compare-data gives an error if the archive is corrupted"""
        self.backup("full", f"{_runtest_dir}/testfiles/various_file_types", options=[])
        output_files = os.listdir(f"{_runtest_dir}/testfiles/output")
        archives = [elem for elem in output_files if "vol" in elem]
        for archive in archives:
            # Edit source file
            with open(f"{_runtest_dir}/testfiles/output/{archive}", "r+") as f:
                f.write("This writes text into each archive file to corrupt it.")
        # Test verify for the file
        try:
            self.verify(
                f"{_runtest_dir}/testfiles/various_file_types/executable",
                file_to_verify="executable",
                options=["--compare-data"],
            )
        except CmdError as e:
            # Should return a 21 error code for "hash mismatch"
            self.assertEqual(e.exit_status, 21, str(e))
        else:
            self.fail("Expected Hash Mismatch Error not thrown")


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import re
import unittest
from unittest.mock import patch

from duplicity import config
from duplicity import manifest
from duplicity import path

from . import UnitTestCase


class VolumeInfoTest(UnitTestCase):
    """Test VolumeInfo"""

    def test_basic(self):
        """Basic VolumeInfoTest"""
        vi = manifest.VolumeInfo()
        vi.set_info(3, (b"hello", b"there"), None, (), None)
        vi.set_hash("MD5", "aoseutaohe")
        s = vi.to_string()
        assert isinstance(s, (bytes, str))
        # print "---------\n%s\n---------" % s
        vi2 = manifest.VolumeInfo()
        vi2.from_string(s)
        assert vi == vi2

    def test_special(self):
        """Test VolumeInfo with special characters"""
        vi = manifest.VolumeInfo()
        vi.set_info(
            3234,
            (
                b"\n eu \x233",
                b"heuo",
                b"\xd8\xab\xb1Wb\xae\xc5]\x8a\xbb\x15v*\xf4\x0f!\xf9>\xe2Y\x86\xbb\xab\xdbp\xb0\x84\x13k\x1d\xc2\xf1\xf5e\xa5U\x82\x9aUV\xa0\xf4\xdf4\xba\xfdX\x03\x82\x07s\xce\x9e\x8b\xb34\x04\x9f\x17 \xf4\x8f\xa6\xfa\x97\xab\xd8\xac\xda\x85\xdcKvC\xfa#\x94\x92\x9e\xc9\xb7\xc3_\x0f\x84g\x9aB\x11<=^\xdbM\x13\x96c\x8b\xa7|*\"\\'^$@#!(){}?+ ~` ",  # noqa
            ),
            None,
            (b"\n",),
            None,
        )
        s = vi.to_string()
        assert isinstance(s, (str, bytes))
        # print "---------\n%s\n---------" % s
        vi2 = manifest.VolumeInfo()
        vi2.from_string(s)
        assert vi == vi2

    def test_contains(self):
        """Test to see if contains() works"""
        vi = manifest.VolumeInfo()
        vi.set_info(1, ("1", "2"), None, ("1", "3"), None)
        assert vi.contains(("1",), recursive=1)
        assert not vi.contains(("1",), recursive=0)

        vi2 = manifest.VolumeInfo()
        vi2.set_info(1, ("A",), None, ("Z",), None)
        assert vi2.contains(("M",), recursive=1)
        assert vi2.contains(("M",), recursive=0)

        vi3 = manifest.VolumeInfo()
        vi3.set_info(1, ("A",), None, ("Z",), None)
        assert not vi3.contains(("3",), recursive=1)
        assert not vi3.contains(("3",), recursive=0)


class ManifestTest(UnitTestCase):
    """Test Manifest class"""

    def setUp(self):
        super().setUp()
        UnitTestCase.setUp(self)
        self.old_files_changed = config.file_changed
        config.file_changed = "testing"

    def tearDown(self):
        config.file_changed = self.old_files_changed
        super().tearDown()

    def test_basic(self):
        vi1 = manifest.VolumeInfo()
        vi1.set_info(3, (b"hello",), None, (), None)
        vi2 = manifest.VolumeInfo()
        vi2.set_info(4, (b"goodbye", b"there"), None, (b"aoeusht",), None)
        vi3 = manifest.VolumeInfo()
        vi3.set_info(34, (), None, (), None)
        m = manifest.Manifest()
        for vi in [vi1, vi2, vi3]:
            m.add_volume_info(vi)

        self.set_config("local_path", path.Path("Foobar"))
        m.set_dirinfo()
        m.set_files_changed_info([])

        s = m.to_string()
        assert s.lower().startswith(b"hostname")
        assert s.endswith(b"\n")

        m2 = manifest.Manifest().from_string(s)
        assert m == m2

    def test_corrupt_filelist(self):
        vi1 = manifest.VolumeInfo()
        vi1.set_info(3, (b"hello",), None, (), None)
        vi2 = manifest.VolumeInfo()
        vi2.set_info(4, (b"goodbye", b"there"), None, (b"aoeusht",), None)
        vi3 = manifest.VolumeInfo()
        vi3.set_info(34, (), None, (), None)
        m = manifest.Manifest()
        for vi in [vi1, vi2, vi3]:
            m.add_volume_info(vi)

        self.set_config("local_path", path.Path("Foobar"))
        m.set_dirinfo()
        m.set_files_changed_info(
            [
                (b"one", b"new"),
                (b"two", b"changed"),
                (b"three", b"new"),
            ]
        )

        # build manifest string
        s = m.to_string()

        # make filecount higher than files in list
        s2 = re.sub(b"Filelist 3", b"Filelist 5", s)
        m2 = manifest.Manifest().from_string(s2)
        assert hasattr(m2, "corrupt_filelist")

    def test_hostname_checks(self):
        self.set_config("hostname", "hostname")
        self.set_config("fqdn", "fqdn")
        m = manifest.Manifest()

        # Matching hostname should work
        m.hostname = "hostname"
        m.check_dirinfo()

        # Matching fqdn should also work for backwards compatibility
        m.hostname = "fqdn"
        m.check_dirinfo()

        # Bad match should throw a fatal error and quit
        m.hostname = "foobar"
        self.assertRaises(SystemExit, m.check_dirinfo)

        # But not if we tell the system to ignore it
        self.set_config("allow_source_mismatch", True)
        m.check_dirinfo()


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2014 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import io
import os
import subprocess
import unittest
import logging

import pytest

import duplicity.backend
from duplicity import log
from duplicity import path
from duplicity import util
from duplicity.errors import BackendException
from testing import _runtest_dir
from . import UnitTestCase


class BackendInstanceBase(UnitTestCase):
    def setUp(self):
        super().setUp()
        assert not os.system(f"rm -rf {_runtest_dir}/testfiles")
        os.makedirs(f"{_runtest_dir}/testfiles")
        self.backend = None
        self.local = path.Path(f"{_runtest_dir}/testfiles/local")
        self.local.writefileobj(io.BytesIO(b"hello"))

    def tearDown(self):
        assert not os.system(f"rm -rf {_runtest_dir}/testfiles")
        if self.backend is None:
            return
        if hasattr(self.backend, "_close"):
            self.backend._close()
        super().tearDown()

    def test_get(self):
        if self.backend is None:
            return
        self.backend._put(self.local, b"file-a")
        getfile = path.Path(f"{_runtest_dir}/testfiles/getfile")
        self.backend._get(b"file-a", getfile)
        self.assertTrue(self.local.compare_data(getfile))

    def test_list(self):
        if self.backend is None:
            return
        self.backend._put(self.local, b"file-a")
        self.backend._put(self.local, b"file-b")
        # It's OK for backends to create files as a side effect of put (e.g.
        # the par2 backend does), so only check that at least a and b exist.
        self.assertTrue(b"file-a" in self.backend._list())
        self.assertTrue(b"file-b" in self.backend._list())

    def test_delete(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, "_delete"):
            self.assertTrue(hasattr(self.backend, "_delete_list"))
            return
        self.backend._put(self.local, b"file-a")
        self.backend._put(self.local, b"file-b")
        self.backend._delete(b"file-a")
        self.assertFalse(b"file-a" in self.backend._list())
        self.assertTrue(b"file-b" in self.backend._list())

    def test_delete_clean(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, "_delete"):
            self.assertTrue(hasattr(self.backend, "_delete_list"))
            return
        self.backend._put(self.local, b"file-a")
        self.backend._delete(b"file-a")
        self.assertFalse(b"file-a" in self.backend._list())

    def test_delete_missing(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, "_delete"):
            self.assertTrue(hasattr(self.backend, "_delete_list"))
            return
        # Backends can either silently ignore this, or throw an error
        # that gives log.ErrorCode.backend_not_found.
        try:
            self.backend._delete(b"file-a")
        except BackendException as e:
            pass  # Something went wrong, but it was an 'expected' something
        except Exception as e:
            code = duplicity.backend._get_code_from_exception(self.backend, "delete", e)
            self.assertEqual(code, log.ErrorCode.backend_not_found)

    def test_delete_list(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, "_delete_list"):
            self.assertTrue(hasattr(self.backend, "_delete"))
            return
        self.backend._put(self.local, b"file-a")
        self.backend._put(self.local, b"file-b")
        self.backend._put(self.local, b"file-c")
        self.backend._delete_list([b"file-a", b"d", b"file-c"])
        files = self.backend._list()
        self.assertFalse(b"file-a" in files, files)
        self.assertTrue(b"file-b" in files, files)
        self.assertFalse(b"file-c" in files, files)

    def test_move(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, "_move"):
            return

        copy = path.Path(f"{_runtest_dir}/testfiles/copy")
        self.local.copy(copy)

        self.backend._move(self.local, b"file-a")
        self.assertTrue(b"file-a" in self.backend._list())
        self.assertFalse(self.local.exists())

        getfile = path.Path(f"{_runtest_dir}/testfiles/getfile")
        self.backend._get(b"file-a", getfile)
        self.assertTrue(copy.compare_data(getfile))

    def test_query_exists(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, "_query"):
            return
        self.backend._put(self.local, b"file-a")
        info = self.backend._query(b"file-a")
        self.assertEqual(info["size"], self.local.getsize())

    def test_query_missing(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, "_query"):
            return
        # Backends can either return -1 themselves, or throw an error
        # that gives log.ErrorCode.backend_not_found.
        try:
            info = self.backend._query(b"missing-file")
        except BackendException as e:  # pylint:
            pass  # Something went wrong, but it was an 'expected' something
        except Exception as e:
            code = duplicity.backend._get_code_from_exception(self.backend, "query", e)
            self.assertEqual(code, log.ErrorCode.backend_not_found)
        else:
            self.assertEqual(info["size"], -1)

    def test_query_list(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, "_query_list"):
            return
        self.backend._put(self.local, b"file-a")
        self.backend._put(self.local, b"file-c")
        info = self.backend._query_list([b"file-a", b"file-b"])
        self.assertEqual(info[b"file-a"]["size"], self.local.getsize())
        self.assertEqual(info[b"file-b"]["size"], -1)
        self.assertFalse(b"file-c" in info)


class LocalBackendTest(BackendInstanceBase):
    def setUp(self):
        super().setUp()
        url = f"file://{_runtest_dir}/testfiles/output"
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, "LocalBackend")


# TODO: Add par2-specific tests here, to confirm that we can recover
@unittest.skipIf(not util.which("par2"), "par2 not installed")
class Par2BackendTest(BackendInstanceBase):
    def setUp(self):
        super().setUp()
        url = f"par2+file://{_runtest_dir}/testfiles/output"
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, "Par2Backend")


# TODO: Fix so localhost is not required.  Fails on LP and GitLab
# class RsyncBackendTest(BackendInstanceBase):
#     def setUp(self):
#         super().setUp()
#         os.makedirs('{0}/testfiles/output')  # rsync needs it to exist first
#         url = 'rsync://localhost:2222//%s/{0}/testfiles/output' % os.getcwd()
#         self.backend = duplicity.backend.get_backend_object(url)
#         self.assertEqual(self.backend.__class__.__name__, 'RsyncBackend')


class TahoeBackendTest(BackendInstanceBase):
    def setUp(self):
        super().setUp()
        os.makedirs(f"{_runtest_dir}/testfiles/output")
        url = f"tahoe://{_runtest_dir}/testfiles/output"
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, "TAHOEBackend")


# TODO: Modernize hsi backend stub
# class HSIBackendTest(BackendInstanceBase):
#     def setUp(self):
#         super().setUp()
#         os.makedirs('{0}/testfiles/output')
#         # hostname is ignored...  Seemingly on purpose
#         url = 'hsi://hostname%s/{0}/testfiles/output' % os.getcwd()
#         self.backend = duplicity.backend.get_backend_object(url)
#         self.assertEqual(self.backend.__class__.__name__, 'HSIBackend')


@unittest.skipIf(not util.which("lftp"), "lftp not installed")
class FTPBackendTest(BackendInstanceBase):
    def setUp(self):
        super().setUp()
        os.makedirs(f"{_runtest_dir}/testfiles/output")
        url = f"ftp://user:pass@hostname/{_runtest_dir}/testfiles/output"
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, "LFTPBackend")


@unittest.skipIf(not util.which("lftp"), "lftp not installed")
class FTPSBackendTest(BackendInstanceBase):
    def setUp(self):
        super().setUp()
        os.makedirs(f"{_runtest_dir}/testfiles/output")
        url = f"ftps://user:pass@hostname/{_runtest_dir}/testfiles/output"
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, "LFTPBackend")


@unittest.skipIf(not util.which("rclone"), "rclone not installed")
class RCloneBackendTest(BackendInstanceBase):
    def setUp(self):
        super().setUp()
        # add a duptest local config
        try:
            assert not os.system("rclone config create duptest local config_is_local true")
            self.delete_config = True
        except Exception as e:
            self.delete_config = False
        os.makedirs(f"{_runtest_dir}/testfiles/output")
        url = f"rclone://duptest:/{_runtest_dir}/testfiles/output"
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, "RcloneBackend")

    def tearDown(self):
        if self.delete_config:
            assert not os.system("rclone config delete duptest")
        super().tearDown()


# TODO: Need fix to work on both Gitlab Docker and personal Docker.
# def in_docker():
#     return os.path.exists("/.dockerenv")
#
#
# TODO: Find out why ssh does not work in Docker.
# @unittest.skipIf(not in_docker(), "Requires Docker / duplicity_test")
# class SFTPBackendTest(BackendInstanceBase):
#     def setUp(self):
#         super().setUp()
#         url = f"pexpect+sftp://testuser:testuser@ssh_server/testdup"
#         self.backend = duplicity.backend.get_backend_object(url)
#         self.assertEqual(self.backend.__class__.__name__, "SSHPxpectBackend")
#         for fn in b"file-a", b"file-b", b"file-c":
#             try:
#                 self.backend._delete(fn)
#             except Exception as e:
#                 log.Error(f"An exception occurred while deleting file {fn}: {e}")
#                 pass



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2011 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import pexpect
import platform
import pytest
import unittest

from duplicity import log
from duplicity.backends._testbackend import BackendErrors as BE
from testing import _runtest_dir
from . import (
    CmdError,
    EnvController,
    FunctionalTestCase,
)


# os.environ['PYDEVD'] = "vscode"


@pytest.mark.usefixtures("redirect_stdin")
class ConcurrencyFullLivecycleTest(FunctionalTestCase):
    def test_verify_compare_data(self):
        """Test that verify works in the basic case when the --compare-data option is used"""
        self.make_largefiles()
        self.backup("full", f"{_runtest_dir}/testfiles/largefiles", options=["--concurrency=3"])

        # Test verify for the file with --compare-data
        self.verify(
            f"{_runtest_dir}/testfiles/largefiles",
            options=["--compare-data"],
        )


@pytest.mark.usefixtures("redirect_stdin")
class ConcurrencyFailTest(FunctionalTestCase):
    def setUp(self):
        super().setUp()

    @pytest.mark.slow
    def test_put_fail_volume(self):
        """
        _testbackend won't put a certain volume
        """
        self.make_largefiles()
        options = [
            "--num-ret=2",
            "--backend-ret=3",
            "--concurrency=3",
        ]
        self.backup_with_failure(
            "full",
            f"{_runtest_dir}/testfiles/largefiles",
            BE.FAIL_WITH_EXCEPTION,
            "vol2.difftar",
            log.ErrorCode.backend_error,
            options=options,
            # PYDEVD="vscode",
            timeout=60,
        )

    @pytest.mark.slow
    def test_put_fail_sys_exit(self):
        """
        _testbackend exit on volume
        """
        self.make_largefiles()
        options = [
            "--num-ret=2",
            "--backend-ret=3",
            "--concurrency=3",
        ]
        self.backup_with_failure(
            "full",
            f"{_runtest_dir}/testfiles/largefiles",
            BE.FAIL_SYSTEM_EXIT,
            "vol2.difftar",
            log.ErrorCode.exception,
            options=options,
            # PYDEVD="vscode",
            timeout=60,
        )

    @pytest.mark.slow
    @unittest.skipIf(
        platform.machine() in ["ppc64el", "ppc64le"],
        "See https://gitlab.com/duplicity/duplicity/-/issues/820",
    )
    def test_out_of_order_volume(self):
        self.make_largefiles()
        options = [
            "--num-ret=2",
            "--backend-ret=3",
            "--concurrency=3",
            "--log-timestamp",
            "--verbosity",
            "i",
        ]
        with EnvController(**{BE.WAIT_FOR_OTHER_VOLUME: '["vol2.difftar", "vol7.difftar"]'}):
            try:
                transferred_files = self.backup("full", f"{_runtest_dir}/testfiles/largefiles", options, timeout=60)
            except pexpect.exceptions.TIMEOUT:
                self.fail(
                    "Concurrent backup was not able to terminate itself. (Most likely caused by a hanging thread.)"
                )
            except CmdError as e:  # Backup muse fail with an exit code != 0
                self.assertEqual(e.exit_status, 0, f"Backup must not fail, because out of order execution. {e}")

    @pytest.mark.slow
    def test_wrong_size(self):
        self.make_largefiles()
        options = [
            "--num-ret=2",
            "--backend-ret=3",
            "--concurrency=3",
            "--log-timestamp",
            "--verbosity",
            "i",
        ]
        with EnvController(**{BE.LAST_BYTE_MISSING: "vol3.difftar"}):
            try:
                transferred_files = self.backup("full", f"{_runtest_dir}/testfiles/largefiles", options, timeout=60)
            except pexpect.exceptions.TIMEOUT:
                self.fail(
                    "Concurrent backup was not able to terminate itself. (Mostlikely caused by a hanging thread.)"
                )
            except CmdError as e:  # Backup muse fail with an exit code != 0
                self.assertEqual(
                    e.exit_status,
                    log.ErrorCode.backend_validation_failed,
                    f"Backup must not fail, because out of order execution. {e}",
                )

    @pytest.mark.slow
    @unittest.skipIf(
        platform.machine() in ["ppc64el", "ppc64le"],
        "Skip on ppc64el or ppc64le machines",
    )
    def test_continue_after_missing_volume(self):
        """
        test recovery after a volume in the sequence is missing.
        vol1, vol3-vol7 should be successful trasferred but val2 fails late.
        """
        # os.environ['PYDEVD'] = ""
        self.make_largefiles()
        options = [
            "--num-ret=2",
            "--backend-ret=3",
            "--concurrency=3",
            "--log-timestamp",
            "--verbosity",
            "i",
        ]
        with EnvController(
            **{BE.WAIT_FOR_OTHER_VOLUME: '["vol2.difftar", "vol5.difftar"]', BE.SKIP_PUT_SILENT: "vol2.difftar"}
        ):
            try:
                transferred_files = self.backup("full", f"{_runtest_dir}/testfiles/largefiles", options, timeout=60)
            except pexpect.exceptions.TIMEOUT:
                self.fail(
                    "Concurrent backup was not able to terminate itself. (Most likely caused by a hanging thread.)"
                )
            except CmdError as e:  # Backup muse fail with an exit code != 0
                self.assertNotEqual(e.exit_status, 0, f"Backup is expected to fail as a volume is missing. {e}")

        try:
            transferred_files = self.backup("full", f"{_runtest_dir}/testfiles/largefiles", options, timeout=60)
        except pexpect.exceptions.TIMEOUT:
            self.fail("Concurrent backup was not able to terminate itself. (Mostlikely caused by a hanging thread.)")
        except CmdError as e:  # Backup muse fail with an exit code != 0
            self.assertEqual(e.exit_status, 0, f"Backup must not fail, because out of order execution. {e}")

        # Test verify for the file with --compare-data
        self.verify(
            f"{_runtest_dir}/testfiles/largefiles",
            options=["--compare-data"],
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
"""py-unit tests for GnuPG

COPYRIGHT:

Copyright (C) 2001  Frank J. Tobin, ftobin@neverending.org

LICENSE:

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
or see http://www.gnu.org/copyleft/lesser.html
"""


import tempfile
import unittest

from duplicity import gpginterface

__author__ = "Frank J. Tobin, ftobin@neverending.org"
__version__ = "0.2.2"
__revision__ = "$Id: GnuPGInterfacetest.py,v 1.11 2009/06/06 17:35:19 loafman Exp $"


class BasicTest(unittest.TestCase):
    """an initializer superclass"""

    def __init__(self, methodName=None):
        self.gnupg = gpginterface.GnuPG()
        unittest.TestCase.__init__(self, methodName)


class GnuPGTests(BasicTest):
    """Tests for GnuPG class"""

    def __init__(self, methodName=None):
        BasicTest.__init__(self, methodName)

        self.gnupg.passphrase = "Three blind mice"
        self.gnupg.options.armor = 1
        self.gnupg.options.meta_interactive = 0
        self.gnupg.options.extra_args.append("--no-secmem-warning")

    def do_create_fh_operation(self, args, input, passphrase=None):  # pylint: disable=redefined-builtin
        creations = ["stdin", "stdout"]

        # Make sure we're getting the passphrase to GnuPG
        # somehow!
        assert passphrase is not None or self.gnupg.passphrase is not None, "No way to send the passphrase to GnuPG!"

        # We'll handle the passphrase manually
        if passphrase is not None:
            creations.append("passphrase")

        proc = self.gnupg.run(args, create_fhs=creations)

        if passphrase is not None:
            proc.handles["passphrase"].write(passphrase)
            proc.handles["passphrase"].close()

        proc.handles["stdin"].write(input)
        proc.handles["stdin"].close()

        ciphertext = proc.handles["stdout"].read()
        proc.handles["stdout"].close()

        # Checking to make sure GnuPG exited successfully
        proc.wait()

        return ciphertext

    def do_attach_fh_operation(self, args, stdin, stdout, passphrase=None):
        # Make sure we're getting the passphrase to GnuPG
        # somehow!
        assert passphrase is not None or self.gnupg.passphrase is not None, "No way to send the passphrase to GnuPG!"

        creations = []
        attachments = {"stdin": stdin, "stdout": stdout}

        proc = self.gnupg.run(args, create_fhs=creations, attach_fhs=attachments)

        # We'll handle the passphrase manually
        if passphrase is not None:
            proc.handles.append("passphrase")

        if passphrase is not None:
            proc.handles["passphrase"].write(passphrase)
            proc.handles["passphrase"].close()

        # Checking to make sure GnuPG exited successfully
        proc.wait()

    def test_create_fhs_solely(self):
        """Do GnuPG operations using solely the create_fhs feature"""
        plaintext = b"Three blind mice"

        ciphertext = self.do_create_fh_operation(["--symmetric"], plaintext)

        decryption = self.do_create_fh_operation(["--decrypt"], ciphertext, self.gnupg.passphrase)
        assert decryption == plaintext, "GnuPG decrypted output does not match original input"

    def test_attach_fhs(self):
        """Do GnuPG operations using the attach_fhs feature"""
        plaintext_source = __file__

        plainfile = open(plaintext_source, "rb")
        temp1 = tempfile.TemporaryFile()
        temp2 = tempfile.TemporaryFile()

        self.do_attach_fh_operation(["--symmetric"], stdin=plainfile, stdout=temp1)

        temp1.seek(0)

        self.do_attach_fh_operation(["--decrypt"], stdin=temp1, stdout=temp2)

        plainfile.seek(0)
        temp2.seek(0)

        assert fh_cmp(plainfile, temp2), "GnuPG decrypted output does not match original input"


class OptionsTests(BasicTest):
    """Tests for Options class"""

    def __init__(self, methodName=None):
        BasicTest.__init__(self, methodName)
        self.reset_options()

    def reset_options(self):
        self.gnupg.options = gpginterface.Options()

    def option_to_arg(self, option):
        return f"--{option.replace('_', '-')}"

    def test_boolean_args(self):
        """test Options boolean options that they generate
        proper arguments"""

        booleans = [
            "armor",
            "no_greeting",
            "no_verbose",
            "batch",
            "always_trust",
            "rfc1991",
            "quiet",
            "openpgp",
            "force_v3_sigs",
            "no_options",
            "textmode",
        ]

        for option in booleans:
            self.reset_options()
            setattr(self.gnupg.options, option, 1)
            arg = self.option_to_arg(option)

            should_be = [arg]
            result = self.gnupg.options.get_args()

            assert (
                should_be == result
            ), f"failure to set option '{option}'; should be {should_be}, but result is {result}"

    def test_string_args(self):
        """test Options string-taking options that they generate
        proper arguments"""

        strings = ["homedir", "default_key", "comment", "compress_algo", "options"]

        string_value = "test-argument"

        for option in strings:
            self.reset_options()
            setattr(self.gnupg.options, option, string_value)
            arg = self.option_to_arg(option)

            should_be = [arg, string_value]
            result = self.gnupg.options.get_args()

            assert (
                should_be == result
            ), f"failure to set option '{option}'; should be {should_be}, but result is {result}"

    def test_list_args(self):
        """test Options string-taking options that they generate
        proper arguments"""

        lists = ["recipients", "encrypt_to"]
        list_value = ["test1", "test2"]

        for option in lists:
            self.reset_options()
            setattr(self.gnupg.options, option, list_value)

            # special case for recipients, since their
            # respective argument is 'recipient', not 'recipients'
            if option == "recipients":
                arg = "--recipient"
            else:
                arg = self.option_to_arg(option)

            should_be = []
            for v in list_value:
                should_be.extend([arg, v])

            result = self.gnupg.options.get_args()

            assert (
                should_be == result
            ), f"failure to set option '{option}'; should be {should_be}, but result is {result}"


class PipesTests(unittest.TestCase):
    """Tests for Pipes class"""

    def test_constructor(self):
        self.pipe = gpginterface.Pipe(1, 2, 0)
        assert self.pipe.parent == 1
        assert self.pipe.child == 2
        assert not self.pipe.direct


########################################################################


def fh_cmp(f1, f2, bufsize=8192):
    while True:
        b1 = f1.read(bufsize)
        b2 = f2.read(bufsize)
        if b1 != b2:
            return 0
        if not b1:
            return 1


########################################################################


if __name__ == "__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2012 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import os
import platform
import sys
import time

import pexpect

from duplicity import backend
from duplicity.backends._testbackend import BackendErrors
from duplicity import config
from duplicity import util
from testing import DuplicityTestCase
from testing import _runtest_dir
from testing import _top_dir


class CmdError(Exception):
    """Indicates an error running an external command"""

    def __init__(self, code):
        Exception.__init__(self, code)
        self.exit_status = code


class FunctionalTestCase(DuplicityTestCase):
    _setsid_w = None

    @classmethod
    def _check_setsid(cls):
        if cls._setsid_w is not None:
            return
        if platform.platform().startswith("Linux"):
            # setsid behavior differs between distributions.
            # If setsid supports -w ("wait"), use it.
            import subprocess

            try:
                with open("/dev/null", "w") as sink:
                    subprocess.check_call(["setsid", "-w", "ls"], stdout=sink, stderr=sink)
            except subprocess.CalledProcessError:
                cls._setsid_w = False
            else:
                cls._setsid_w = True

    def setUp(self):
        super().setUp()

        self.unpack_testfiles()

        self.class_args = []
        self.backend_url = f"fortestsonly://{_runtest_dir}/testfiles/output"
        self.last_backup = None
        self.set_environ("PASSPHRASE", self.sign_passphrase)
        self.set_environ("SIGN_PASSPHRASE", self.sign_passphrase)

        backend_inst = backend.get_backend(self.backend_url)
        bl = backend_inst.list()
        if bl:
            backend_inst.delete(backend_inst.list())
        backend_inst.close()
        self._check_setsid()

    def run_duplicity(self, options=None, current_time=None, fail=None, passphrase_input=None, timeout=None):
        """
        Run duplicity binary with given arguments and options
        """

        if options is None:
            options = []
        if passphrase_input is None:
            passphrase_input = []

        # Check all string inputs are unicode -- we will convert to system encoding before running the command
        for item in passphrase_input:
            assert isinstance(item, str), f"item {os.fsdecode(item)} in passphrase_input is not unicode"

        # set python path to be dev directory
        os.environ["PYTHONPATH"] = _top_dir

        cmd_list = []

        # We run under setsid and take input from /dev/null (below) because
        # this way we force a failure if duplicity tries to read from the
        # console unexpectedly (like for gpg password or such).
        if platform.platform().startswith("Linux"):
            cmd_list.extend(["setsid"])
            if self._setsid_w:
                cmd_list.extend(["-w"])

        cmd_list.extend([f"python{sys.version_info.major}.{sys.version_info.minor}"])

        if os.environ.get("RUN_COVERAGE", None):
            cmd_list.extend(["-m", "coverage", "run", "--source=duplicity", "-p"])

        cmd_list.extend([os.path.join(_top_dir, "duplicity", "__main__.py")])
        cmd_list.extend(options)

        if os.environ.get("PYDEVD", None):
            cmd_list.extend(["--pydevd"])

        if os.environ.get("TESTDEBUG", False):
            # enable debug logging for trouble shooting
            cmd_list.extend(["-vd"])
        else:
            # keep duplicity quite for a condensed test log.
            cmd_list.extend(["-v0"])
        cmd_list.extend(["--no-print-statistics"])
        cmd_list.extend([f"--archive-dir={_runtest_dir}/testfiles/cache"])

        if current_time:
            cmd_list.extend(["--current-time", current_time])

        cmd_list.extend(self.class_args)

        dup_env = dict(os.environ)
        if fail:
            # cmd_list.extend(["--fail", fail]) replaced by ENV
            dup_env[BackendErrors.FAIL_SYSTEM_EXIT] = f"vol{fail}.difftar"
            cmd_list.extend(["--num-ret=2", "--backend-ret=3"])  # fail faster

        # convert to string and single quote to avoid shell expansion
        cmdline = " ".join([f"'{x}'" for x in cmd_list])

        if not passphrase_input:
            cmdline += " < /dev/null"

        # Set encoding to filesystem encoding and send to spawn.
        # option -f to shell avoids globbing which breaks selection tests.
        # shell used instead of cmdline to allow setsid to be prepended.
        child = pexpect.spawn(
            "/bin/sh",
            ["-f", "-c", cmdline],
            timeout=timeout,
            env=dup_env,  # type: ignore
            encoding=config.fsencoding,
        )

        for passphrase in passphrase_input:
            child.expect("passphrase.*:")
            child.sendline(passphrase)

        # if the command fails, we need to clear its output
        # so it will terminate cleanly.
        child.expect_exact(pexpect.EOF)
        lines = child.before.splitlines()
        child.wait()
        child.ptyproc.delayafterclose = 0.0
        return_val = child.exitstatus

        print("\n...command:", cmdline, file=sys.stderr)
        print("...cwd:", os.getcwd(), file=sys.stderr)
        print("...output:", file=sys.stderr)
        for line in lines:
            line = line.rstrip()
            if line:
                print(os.fsdecode(line), file=sys.stderr)
        print("...return_val:", return_val, file=sys.stderr)
        if fail:
            self.assertEqual(30, return_val)
        elif return_val:
            raise CmdError(return_val)

    def backup(self, type, input_dir, options=None, **kwargs):  # pylint: disable=redefined-builtin
        """Run duplicity backup to default directory"""
        if options is None:
            options = []
        options = [type, input_dir, self.backend_url, "--volsize=1"] + options
        before_files = self.get_backend_files()

        # If a chain ends with time X and the next full chain begins at time X,
        # we may trigger an assert in dup_collections.py.  If needed, sleep to
        # avoid such problems
        now = time.time()
        if self.last_backup == int(now):
            time.sleep(1)

        self.run_duplicity(options=options, **kwargs)
        self.last_backup = int(time.time())

        after_files = self.get_backend_files()
        return after_files - before_files

    def backup_with_failure(
        self, type, input_dir, failure_type, failure_condition, error_code, options=None, PYDEVD=None, **kwargs
    ):
        """
        using _testbackent to trigger certain failure conditions. See backends/_testbackend.py for possible trigger
        """
        if not options:
            options = [  # lower the retry count to fail faster.
                "--num-ret=2",
                "--backend-ret=3",
            ]

        try:
            env = {failure_type: failure_condition}
            if PYDEVD:  # start debugger in forked duplicity execution. Use for troubleshooting only.
                env["PYDEVD"] = PYDEVD
            with EnvController(**env):
                self.backup(type, input_dir, options, **kwargs)
        except CmdError as e:  # Backup must fail with an exit code != 0
            self.assertEqual(e.exit_status, error_code, str(e))
        else:
            self.fail("Expected CmdError not thrown")

    def restore(self, file_to_restore=None, time=None, options=None, **kwargs):
        if options is None:
            options = []
        assert not os.system(f"rm -rf {_runtest_dir}/testfiles/restore_out")
        options = [
            "restore",
            self.backend_url,
            f"{_runtest_dir}/testfiles/restore_out",
        ] + options
        if file_to_restore:
            options.extend(["--path-to-restore", file_to_restore])
        if time:
            options.extend(["--restore-time", str(time)])
        self.run_duplicity(options=options, **kwargs)

    def verify(self, dirname, file_to_verify=None, time=None, options=None, **kwargs):
        if options is None:
            options = []
        options = ["verify", self.backend_url, dirname] + options
        if file_to_verify:
            options.extend(["--path-to-restore", file_to_verify])
        if time:
            options.extend(["--restore-time", str(time)])
        self.run_duplicity(options=options, **kwargs)

    def cleanup(self, options=None):
        """
        Run duplicity cleanup to default directory
        """
        if options is None:
            options = []
        options = ["cleanup", self.backend_url, "--force"] + options
        self.run_duplicity(options=options)

    def collection_status(self, options=None):
        """
        Run duplicity collection-status to default directory
        """
        if options is None:
            options = []
        options = ["collection-status", self.backend_url] + options
        self.run_duplicity(options=options)

    def get_backend_files(self):
        backend_inst = backend.get_backend(self.backend_url)
        bl = backend_inst.list()
        backend_inst.close()
        return set(bl)

    def make_largefiles(self, count=3, size=2):
        """
        Makes a number of large files in /tmp/testfiles/largefiles that each are
        the specified number of megabytes.
        """
        assert not os.system(f"mkdir {_runtest_dir}/testfiles/largefiles")
        for n in range(count):
            assert not os.system(
                f"dd if=/dev/urandom of={_runtest_dir}/testfiles/largefiles/file{n + 1} "
                f"bs=1024 count={size * 1024} > /dev/null 2>&1"
            )


class EnvController:
    def __init__(self, **kwargs):
        self.env = kwargs

    def __enter__(self):
        for k, v in self.env.items():
            os.environ[k] = v

    def __exit__(self, exc_type, exc_val, exc_tb):
        for k, _ in self.env.items():
            os.unsetenv(k)
            del os.environ[k]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2008 Michael Terry <mike@mterry.name>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


import os
import sys
import unittest

from testing import _runtest_dir
from testing import _top_dir
from . import FunctionalTestCase


class LogTest(FunctionalTestCase):
    """Test machine-readable functions/classes in log.py"""

    logfile = f"{_runtest_dir}/duplicity.log"

    def setUp(self):
        super().setUp()
        assert not os.system(f"rm -f {self.logfile}")

    def tearDown(self):
        assert not os.system(f"rm -f {self.logfile}")
        super().tearDown()

    def test_command_line_error(self):
        """Check notification of a simple error code"""

        # Run actual duplicity command (will fail because bad dirs passed)
        basepython = f"python{sys.version_info.major}.{sys.version_info.minor}"
        cmd = (
            f"{basepython} {_top_dir}/duplicity/__main__.py --log-file={self.logfile} "
            f"full testing baddir >/dev/null 2>&1"
        )
        os.system(cmd)

        # The format of the file should be:
        # """ERROR 23 CommandLineError
        # . Blah blah blah.
        # . Blah blah blah.
        #
        # """
        f = open(self.logfile, "r")
        linecount = 0
        lastline = False
        for line in f:
            assert not lastline
            linecount += 1
            if linecount == 1:
                assert line == "ERROR 23 CommandLineError\n"
            elif line[0] != "\n":
                assert line.startswith(r". ")
            else:
                lastline = True
        assert lastline


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os
import glob
import re
import shutil
import subprocess
import sys
import time
import warnings

warnings.filterwarnings("ignore", message="setup.py install is deprecated")
warnings.filterwarnings("ignore", message="easy_install command is deprecated")
warnings.filterwarnings("ignore", message="pyproject.toml does not contain a tool.setuptools_scm section")
warnings.filterwarnings("ignore", message="Configuring installation scheme with distutils config files")

from setuptools import setup, Extension, Command
from setuptools.command.build_ext import build_ext

# check that we can function here
if not ((3, 8) <= sys.version_info[:2]):
    print("Sorry, duplicity requires version 3.8 thru 3.13 of Python.", file=sys.stderr)
    sys.exit(1)

Version: str = "3.0.4a"
reldate: str = time.strftime("%B %d, %Y", time.gmtime(int(os.environ.get("SOURCE_DATE_EPOCH", time.time()))))

# READTHEDOCS uses setup.py sdist but can't handle extensions
ext_modules = list()
incdir_list = list()
libdir_list = list()
if os.environ.get("READTHEDOCS", None) is None:
    # set incdir and libdir for librsync
    if os.name == "posix":
        LIBRSYNC_DIR = os.environ.get("LIBRSYNC_DIR", "")
        args = sys.argv[:]
        for arg in args:
            if arg.startswith("--librsync-dir="):
                LIBRSYNC_DIR = arg.split("=")[1]
                sys.argv.remove(arg)
        if LIBRSYNC_DIR:
            incdir_list.append(os.path.join(LIBRSYNC_DIR, "include"))
            libdir_list.append(os.path.join(LIBRSYNC_DIR, "lib"))

    # set incdir and libdir for pyenv
    if pyenv_root := os.environ.get("PYENV_ROOT", None):
        major, minor, patch = sys.version_info[:3]
        incdir_list.append(
            os.path.join(
                f"{pyenv_root}",
                f"versions",
                f"{major}.{minor}.{patch}",
                f"include",
                f"python{major}.{minor}",
            )
        )
        libdir_list.append(
            os.path.join(
                f"{pyenv_root}",
                f"versions",
                f"{major}.{minor}.{patch}",
                f"lib",
                f"python{major}.{minor}",
            )
        )

    # add standard locs
    incdir_list.append("/usr/local/include")
    libdir_list.append("/usr/local/lib")
    incdir_list.append("/usr/include")
    libdir_list.append("/usr/lib")

    # build the librsync extension
    ext_modules = [
        Extension(
            name=r"duplicity._librsync",
            sources=["duplicity/_librsyncmodule.c"],
            include_dirs=incdir_list,
            library_dirs=libdir_list,
            libraries=["rsync"],
        )
    ]


def get_data_files():
    """gen list of data files"""

    # static data files
    data_files = [
        (
            "share/man/man1",
            [
                "man/duplicity.1",
            ],
        ),
        (
            f"share/doc/duplicity-{Version}",
            [
                "CHANGELOG.md",
                "AUTHORS.md",
                "COPYING",
                "README.md",
                "README-LOG.md",
                "README-REPO.md",
                "README-TESTING.md",
            ],
        ),
    ]

    # short circuit fot READTHEDOCS
    if os.environ.get("READTHEDOCS") == "True":
        return data_files

    # msgfmt the translation files
    assert os.path.exists("po"), "Missing 'po' directory."

    linguas = glob.glob("po/*.po")
    for lang in linguas:
        lang = lang[3:-3]
        try:
            os.mkdir(os.path.join("po", lang))
        except os.error:
            pass
        subprocess.run(f"cp po/{lang}.po po/{lang}", shell=True, check=True)
        subprocess.run(f"msgfmt po/{lang}.po -o po/{lang}/duplicity.mo", shell=True, check=True)

    for root, dirs, files in os.walk("po"):
        for file in files:
            path = os.path.join(root, file)
            if path.endswith("duplicity.mo"):
                lang = os.path.split(root)[-1]
                data_files.append((f"share/locale/{lang}/LC_MESSAGES", [f"po/{lang}/duplicity.mo"]))

    return data_files


def cleanup():
    if os.path.exists("po/LINGUAS"):
        linguas = open("po/LINGUAS").readlines()
        for line in linguas:
            langs = line.split()
            for lang in langs:
                shutil.rmtree(os.path.join("po", lang), ignore_errors=True)


class BuildExtCommand(build_ext):
    """Build extension modules."""

    def run(self):
        # build the _librsync.so module
        print("Building extension for librsync...")
        self.inplace = True
        build_ext.run(self)


class SetVersionCommand(Command):
    """
    Mod the versioned files and add correct version and reldate
    """

    description: str = "Version source based env var DUP_VERSION"

    user_options: list = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        global Version

        if not (Version := os.environ.get("DUP_VERSION", False).strip("\"'")):
            print("DUP_VERSION not set in environment.\nSet DUP_VERSION and try again")
            sys.exit(1)

        if self.dry_run:
            print("Dry run, no changes will be made.")

        # .TH DUPLICITY 1 "$reldate" "Version $version" "User Manuals" \"  -*- nroff -*-
        self.version_source(
            r"""\.TH\ DUPLICITY\ 1\ "(?P<reldate>[^"]*)"\ "Version\ (?P<version>[^"]*)"\ "User\ Manuals"\ \\"\ """
            r"""\ \-\*\-\ nroff\ \-\*\-""",
            r"""\.TH\ DUPLICITY\ 1\ "(?P<reldate>[^"]*)"\ "Version\ (?P<version>[^"]*)"\ "User\ Manuals"\ \\"\ """
            r"""\ \-\*\-\ nroff\ \-\*\-""",
            os.path.join("man", "duplicity.1"),
        )

        # __version__ = "$version"
        self.version_source(
            r'__version__: str = "(?P<version>[^"]*)"',
            r'__reldate__: str = "(?P<reldate>[^"]*)"',
            os.path.join("duplicity", "__init__.py"),
        )

        # version: $version
        self.version_source(
            r"version: (?P<version>.*)\n",
            None,
            os.path.join("snap", "snapcraft.yaml"),
        )

        # Version: str = "$version"
        self.version_source(
            r'Version: str = "(?P<version>[^\"]*)"',
            None,
            os.path.join(".", "setup.py"),
        )

        # version = "$version"
        self.version_source(
            r'version = "(?P<version>[^\"]*)"',
            None,
            os.path.join(".", "pyproject.toml"),
        )

    def version_source(self, version_patt: str, reldate_patt: str, pathname: str):
        """
        Copy source to dest, substituting current version with Version
        current release date with today's date, i.e. December 28, 2008.
        """
        with open(pathname, "rt") as fd:
            buffer = fd.read()

        # process version
        if version_patt:
            if m := re.search(version_patt, buffer):
                version_sub = re.escape(m.group("version"))
                newbuffer = re.sub(version_sub, Version, buffer)
                if newbuffer == buffer:
                    print(f"ERROR: version unchanged in {pathname}.", file=sys.stderr)
                else:
                    buffer = newbuffer
                    if self.verbose:
                        print(f"Substituted '{version_sub}' with '{Version}' in {pathname}.")
            else:
                print(f"ERROR: {version_patt} not found in {pathname}.", file=sys.stderr)
                sys.exit(1)

        # process reldate
        if reldate_patt:
            if m := re.search(reldate_patt, buffer):
                reldate_sub = re.escape(m.group("reldate"))
                newbuffer = re.sub(reldate_sub, reldate, buffer)
                if newbuffer == buffer:
                    print(f"ERROR: reldate unchanged in {pathname}.", file=sys.stderr)
                else:
                    buffer = newbuffer
                    if self.verbose:
                        print(f"Substituted '{reldate_sub}' with '{reldate}' in {pathname}.")
            else:
                print(f"ERROR: {reldate_patt} not found in {pathname}.", file=sys.stderr)
                sys.exit(1)

        if not self.dry_run:
            with open(pathname, "w") as fd:
                fd.write(buffer)


setup(
    packages=[
        "duplicity",
        "duplicity.backends",
        "duplicity.backends.pyrax_identity",
    ],
    package_dir={
        "duplicity": "duplicity",
        "duplicity.backends": "duplicity/backends",
    },
    ext_modules=ext_modules,
    data_files=get_data_files(),
    include_package_data=True,
    cmdclass={
        "build_ext": BuildExtCommand,
        "setversion": SetVersionCommand,
    },
)

cleanup()
